%% Implementation of a Gauss-Neston SQP solver usign CasADi

close all
clear variables
clc

% import CasADi
import casadi.*

nv = 2;
x = MX.sym('x',nv);

% Define objective function
f = 1/2*(x(1) -1)^2 + 1/2*(10*(x(2) - x(1)^2))^2 + 1/2*x(2)^2;
Jf = Function('Jf', {x},{jacobian(f,x)});

r = [(x(1) -1);(10*(x(2) - x(1)^2)); x(2)];
R = Function('R', {x},{r});
Jr = Function('Jr', {x},{jacobian(r,x)});


% Define constraints
g = x(1) + (1-x(2))^2;
G = Function('G',{x},{g});
Jg = Function('Jg', {x},{jacobian(g,x)});

% Define linearization point
xk = MX.sym('xk',nv);

% Define linearized constraint 
Jtemp = Jg({xk});
g_temp = G({xk});
g_l = g_temp{1} + Jtemp{1}*x;

% Gauss-Newton Hessian approximation
j_out = Jr({xk});
jf_out = Jf({xk});
r_out  = R({xk});
f_gn = 1/2*x'*j_out{1}'*j_out{1}*x + jf_out{1}*x;

% Allocate QP solver
qp = struct('x',x, 'f',f_gn,'g',g_l,'p',xk);
solver = qpsol('solver', 'qpoases', qp);

% SQP solver
max_it = 100;
xk = [1 1]'; % Initial guess
iter = zeros(nv,max_it);
iter(:,1) = xk;

for i=2:max_it    
    
    % Formulate the QP
    arg     = struct;
    arg.lbg =  0;           
    arg.ubg =  0;
    arg.lbx =  -Inf;           
    arg.ubx =  Inf; 
    arg.p   =  xk;

    % Solve with IPOPT
    sol = solver(arg);
    step = full(sol.x);
    if norm(step) < 1.0e-16
        break;
    end
    iter(:,i) = iter(:,i-1) + step;
    xk = iter(:,i);
    
end
iter = iter(:,1:i-1);
[X,Y] = meshgrid(-1.5:.05:1.5, -1.5:.05:1.5);
Z = log(1 + 1/2*(X -1).^2 + 1/2*(10*(Y -X.^2)).^2 + 1/2*Y.^2);
figure()
subplot(1,2,1)
surf(X,Y,Z)
xlabel('x_1')
ylabel('x_2')
hold all
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'black')
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'blacko')
y_g = linspace(-0.08,1.1,20);
x_g = -(1 - y_g).^2;
plot3(x_g,y_g,zeros(length(x_g),1),'r');

x_h = linspace(-1.5,1.5,20);
y_h = 0.2 + x_h.^2;
plot3(x_h,y_h,zeros(length(x_h),1),'r--');

xlim([-1.5,1.5]);
ylim([-1.5,1.5]);
subplot(1,2,2)
contour(X,Y,Z)
hold all
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'black')
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'blacko')
xlabel('x_1')
ylabel('x_2')
plot3(x_g,y_g,zeros(length(x_g),1),'r');
plot3(x_h,y_h,zeros(length(x_h),1),'r--');
xlim([-1.5,1.5]);
ylim([-1.5,1.5]);
figure()
plot(iter(1:2,:)')
grid on
xlabel('iterations')
ylabel('primal solution')
grid on