## Implementation of a Gauss-Neston SQP solver usign CasADi

# import CasADi
import casadi as C
import numpy as np

nv = 2;
x = C.MX.sym('x',nv);

# Define objective function
f = 1.0/2*(x[0] -1)**2 + 1.0/2*(10*(x[1] - x[0]**2))**2 + 1.0/2*x[1]**2
Jf = C.Function('Jf', [x],[C.jacobian(f,x)])

r = C.vertcat([(x[0] -1),(10*(x[1] - x[0]**2)), x[1]])
R = C.Function('R', [x],[r])
Jr = C.Function('Jr', [x],[C.jacobian(r,x)])


# Define constraints
g = x[0] + (1-x[1])**2
G = C.Function('G',[x],[g])
Jg = C.Function('Jg',[x],[C.jacobian(g,x)])

# Define linearization point
xk = C.MX.sym('xk',nv)

# Define linearized constraint 
Jtemp = Jg([xk])
g_temp = G([xk])
g_l = g_temp[0] + C.mtimes(Jtemp[0],x)

# Gauss-Newton Hessian approximation
j_out = Jr([xk])
jf_out = Jf([xk])
r_out  = R([xk])
f_gn = 1.0/2*C.mtimes([x.T,j_out[0].T,j_out[0],x]) + C.mtimes(jf_out[0],x)

# Allocate QP solver
qp = {'x':x, 'f':f_gn,'g':g_l,'p':xk}
solver = C.qpsol('solver', 'qpoases', qp)

# SQP solver
max_it = 100
xk = np.array([1,1]) # Initial guess
iters = [xk]

for i in range(1,max_it):
    
    # Formulate the QP
    arg     = {};
    arg["lbg"] =  0    
    arg["ubg"] =  0
    arg["lbx"] =  -np.Inf      
    arg["ubx"] =  np.Inf
    arg["p"]   =  iters[-1]

    # Solve with IPOPT
    sol = solver(arg);
    step = sol["x"]
    if np.linalg.norm(step) < 1.0e-16:
        break
    iters.append(iters[-1]+step)

[X,Y] = np.meshgrid(np.arange(-1.5,1.5,0.05), np.arange(-1.5,1.5,0.05))
Z = np.log(1 + 1.0/2*(X -1)**2 + 1.0/2*(10*(Y -X**2))**2 + 1.0/2*Y**2)

import pylab
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

pylab.figure()
pylab.subplot(1,2,1,projection='3d')
pylab.gca().plot_surface(X,Y,Z)
pylab.xlabel('x_1')
pylab.ylabel('x_2')

iters = np.array(C.horzcat(iters))


pylab.gca().plot(iters[0,:],iters[1,:],'k')
pylab.gca().plot(iters[0,:],iters[1,:],'ko')
y_g = np.linspace(-0.08,1.1,20)
x_g = -(1 - y_g)**2

pylab.gca().plot(x_g,y_g,'r')

x_h = np.linspace(-1.5,1.5,20)
y_h = 0.2 + x_h**2
pylab.gca().plot(x_h,y_h,'r--')

pylab.subplot(1,2,2)

pylab.contour(X,Y,Z)

pylab.plot(iters[0,:],iters[1,:],'k')
pylab.plot(iters[0,:],iters[1,:],'ko')
y_g = np.linspace(-0.08,1.1,20)
x_g = -(1 - y_g)**2

pylab.plot(x_g,y_g,'r')

x_h = np.linspace(-1.5,1.5,20)
y_h = 0.2 + x_h**2
pylab.plot(x_h,y_h,'r')

pylab.figure()
pylab.plot(iters.T)
pylab.xlabel('iterations')
pylab.ylabel('primal solution')

pylab.grid(True)

pylab.show()
