%% Implementation of a Gauss-Neston SQP solver usign CasADi

close all
clear variables
clc

% import CasADi
import casadi.*

nv = 2;
x = MX.sym('x',nv);

% Define objective function
f = 1/2*(x(1) -1)^2 + 1/2*(10*(x(2) - x(1)^2))^2 + 1/2*x(2)^2;
F = Function('F',{x},{f});
Jf = Function('Jf', {x},{jacobian(f,x)});
x_test = [1.1 3.2];
test_c = F({x_test});
test = full(test_c{1})
test_c = Jf({x_test});
test = full(test_c{1})

r = [(x(1) -1);(10*(x(2) - x(1)^2)); x(2)];
R = Function('R', {x},{r});
Jr = Function('Jr', {x},{jacobian(r,x)});
test_c = R({x_test});
test = full(test_c{1})
test_c = Jr({x_test});
test = full(test_c{1})

% Define equalities 
g = x(1) + (1-x(2))^2;
G = Function('G',{x},{g});
Jg = Function('Jg', {x},{jacobian(g,x)});
test_c = G({x_test});
test = full(test_c{1})
test_c = Jg({x_test});
test = full(test_c{1})


% Define inequalities 
h = 0.2 + x(1)^2-x(2);
H = Function('H',{x},{h});
Jh = Function('Jh', {x},{jacobian(h,x)});
test_c = H({x_test});
test = full(test_c{1})
test_c = Jh({x_test});
test = full(test_c{1})

% Define linearization point
xk = MX.sym('xk',nv);

% Define linearized constraints 
Jtemp = Jg({xk});
g_temp = G({xk});
g_l = g_temp{1} + Jtemp{1}*x;

Jtemp = Jh({xk});
h_temp = H({xk});
h_l = h_temp{1} + Jtemp{1}*x;

% Gauss-Newton Hessian approximation
j_out = Jr({xk});
jf_out = Jf({xk});
r_out  = R({xk});
f_gn = 1/2*x'*j_out{1}'*j_out{1}*x + jf_out{1}*x;

% Allocate QP solver
qp = struct('x',x, 'f',f_gn,'g',[g_l;h_l],'p',xk);
solver = qpsol('solver', 'qpoases', qp);

% SQP solver
max_it = 100;
xk = [1 1]'; % Initial guess
iter = zeros(nv,max_it);
iter(:,1) = xk;

for i=2:max_it    
    
    % Formulate the QP
    arg     = struct;
    arg.lbg =  [0;-Inf];           
    arg.ubg =  [0;0];
    arg.lbx =  -Inf;           
    arg.ubx =  Inf; 
    arg.p   =  xk;

    % Solve with qpOASES
    sol = solver(arg);
    step = full(sol.x);
    if norm(step) < 1.0e-16
        break;
    end
    
    % Line-search 
    t = 1;
    kappa = 0.7;
    alpha = 1.1;
    out = F({iter(:,i-1)});
    prev_cost = full(out{1});
    next_cost = Inf;
    while(next_cost > alpha*prev_cost)
        trial = iter(:,i-1) + t*step;
        out = F({trial});
        next_cost = full(out{1});
        t = t*kappa;
    end
    iter(:,i) = iter(:,i-1) + t*step;
    xk = iter(:,i);
    
end
iter = iter(:,1:i-1);
[X,Y] = meshgrid(-1.5:.05:1.5, -1.5:.05:1.5);
Z = log(1 + 1/2*(X -1).^2 + 1/2*(10*(Y -X.^2)).^2 + 1/2*Y.^2);
figure()
subplot(1,2,1)
surf(X,Y,Z)
xlabel('x_1')
ylabel('x_2')
hold all
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'black')
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'blacko')
y_g = linspace(-0.08,1.1,20);
x_g = -(1 - y_g).^2;
plot3(x_g,y_g,zeros(length(x_g),1),'r');

x_h = linspace(-1.5,1.5,20);
y_h = 0.2 + x_h.^2;
plot3(x_h,y_h,zeros(length(x_h),1),'r--');

xlim([-1.5,1.5]);
ylim([-1.5,1.5]);
subplot(1,2,2)
contour(X,Y,Z)
hold all
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'black')
plot3(iter(1,:),iter(2,:),zeros(length(iter(1,:)),1),'blacko')
xlabel('x_1')
ylabel('x_2')
plot3(x_g,y_g,zeros(length(x_g),1),'r');
plot3(x_h,y_h,zeros(length(x_h),1),'r--');
xlim([-1.5,1.5]);
ylim([-1.5,1.5]);
figure()
plot(iter(1:2,:)')
grid on
xlabel('iterations')
ylabel('primal solution')
grid on